/* This file is auto-generated by sync-with-kernel.sh */

#define CMD32_DMX_ADD_PID 0x40026f33
#define CMD32_DMX_DQBUF 0xc0186f40
#define CMD32_DMX_EXPBUF 0xc00c6f3e
#define CMD32_DMX_GET_PES_PIDS 0x800a6f2f
#define CMD32_DMX_GET_STC 0xc0106f32
#define CMD32_DMX_QBUF 0xc0186f3f
#define CMD32_DMX_QUERYBUF 0xc0186f3d
#define CMD32_DMX_REMOVE_PID 0x40026f34
#define CMD32_DMX_REQBUFS 0xc0086f3c
#define CMD32_DMX_SET_BUFFER_SIZE 0x6f2d
#define CMD32_DMX_SET_FILTER 0x403c6f2b
#define CMD32_DMX_SET_PES_FILTER 0x40146f2c
#define CMD32_DMX_START 0x6f29
#define CMD32_DMX_STOP 0x6f2a
#define CMD32_FE_DISEQC_RECV_SLAVE_REPLY 0x800c6f40
#define CMD32_FE_DISEQC_RESET_OVERLOAD 0x6f3e
#define CMD32_FE_DISEQC_SEND_BURST 0x6f41
#define CMD32_FE_DISEQC_SEND_MASTER_CMD 0x40076f3f
#define CMD32_FE_DISHNETWORK_SEND_LEGACY_CMD 0x6f50
#define CMD32_FE_ENABLE_HIGH_LNB_VOLTAGE 0x6f44
#define CMD32_FE_GET_EVENT 0x80286f4e
#define CMD32_FE_GET_FRONTEND 0x80246f4d
#define CMD32_FE_GET_INFO 0x80a86f3d
#define CMD32_FE_GET_PROPERTY 0x80086f53
#define CMD32_FE_READ_BER 0x80046f46
#define CMD32_FE_READ_SIGNAL_STRENGTH 0x80026f47
#define CMD32_FE_READ_SNR 0x80026f48
#define CMD32_FE_READ_STATUS 0x80046f45
#define CMD32_FE_READ_UNCORRECTED_BLOCKS 0x80046f49
#define CMD32_FE_SET_FRONTEND 0x40246f4c
#define CMD32_FE_SET_FRONTEND_TUNE_MODE 0x6f51
#define CMD32_FE_SET_PROPERTY 0x40086f52
#define CMD32_FE_SET_TONE 0x6f42
#define CMD32_FE_SET_VOLTAGE 0x6f43
#define CMD32_VIDIOC_CREATE_BUFS 0xc0f8565c
#define CMD32_VIDIOC_CROPCAP 0xc02c563a
#define CMD32_VIDIOC_DBG_G_CHIP_INFO 0xc0c85666
#define CMD32_VIDIOC_DBG_G_REGISTER 0xc0385650
#define CMD32_VIDIOC_DBG_S_REGISTER 0x4038564f
#define CMD32_VIDIOC_DECODER_CMD 0xc0485660
#define CMD32_VIDIOC_DQBUF 0xc0445611
#define CMD32_VIDIOC_DQEVENT 0x80785659
#define CMD32_VIDIOC_DV_TIMINGS_CAP 0xc0905664
#define CMD32_VIDIOC_ENCODER_CMD 0xc028564d
#define CMD32_VIDIOC_ENUMAUDIO 0xc0345641
#define CMD32_VIDIOC_ENUMAUDOUT 0xc0345642
#define CMD32_VIDIOC_ENUMINPUT 0xc04c561a
#define CMD32_VIDIOC_ENUMOUTPUT 0xc0485630
#define CMD32_VIDIOC_ENUMSTD 0xc0405619
#define CMD32_VIDIOC_ENUM_DV_TIMINGS 0xc0945662
#define CMD32_VIDIOC_ENUM_FMT 0xc0405602
#define CMD32_VIDIOC_ENUM_FRAMEINTERVALS 0xc034564b
#define CMD32_VIDIOC_ENUM_FRAMESIZES 0xc02c564a
#define CMD32_VIDIOC_ENUM_FREQ_BANDS 0xc0405665
#define CMD32_VIDIOC_EXPBUF 0xc0405610
#define CMD32_VIDIOC_G_AUDIO 0x80345621
#define CMD32_VIDIOC_G_AUDOUT 0x80345631
#define CMD32_VIDIOC_G_CROP 0xc014563b
#define CMD32_VIDIOC_G_CTRL 0xc008561b
#define CMD32_VIDIOC_G_DV_TIMINGS 0xc0845658
#define CMD32_VIDIOC_G_EDID 0xc0245628
#define CMD32_VIDIOC_G_ENC_INDEX 0x8818564c
#define CMD32_VIDIOC_G_EXT_CTRLS 0xc0185647
#define CMD32_VIDIOC_G_FBUF 0x802c560a
#define CMD32_VIDIOC_G_FMT 0xc0cc5604
#define CMD32_VIDIOC_G_FREQUENCY 0xc02c5638
#define CMD32_VIDIOC_G_INPUT 0x80045626
#define CMD32_VIDIOC_G_JPEGCOMP 0x808c563d
#define CMD32_VIDIOC_G_MODULATOR 0xc0445636
#define CMD32_VIDIOC_G_OUTPUT 0x8004562e
#define CMD32_VIDIOC_G_PARM 0xc0cc5615
#define CMD32_VIDIOC_G_PRIORITY 0x80045643
#define CMD32_VIDIOC_G_SELECTION 0xc040565e
#define CMD32_VIDIOC_G_SLICED_VBI_CAP 0xc0745645
#define CMD32_VIDIOC_G_STD 0x80085617
#define CMD32_VIDIOC_G_TUNER 0xc054561d
#define CMD32_VIDIOC_LOG_STATUS 0x5646
#define CMD32_VIDIOC_OVERLAY 0x4004560e
#define CMD32_VIDIOC_PREPARE_BUF 0xc044565d
#define CMD32_VIDIOC_QBUF 0xc044560f
#define CMD32_VIDIOC_QUERYBUF 0xc0445609
#define CMD32_VIDIOC_QUERYCAP 0x80685600
#define CMD32_VIDIOC_QUERYCTRL 0xc0445624
#define CMD32_VIDIOC_QUERYMENU 0xc02c5625
#define CMD32_VIDIOC_QUERYSTD 0x8008563f
#define CMD32_VIDIOC_QUERY_DV_TIMINGS 0x80845663
#define CMD32_VIDIOC_QUERY_EXT_CTRL 0xc0e85667
#define CMD32_VIDIOC_REMOVE_BUFS 0xc0405668
#define CMD32_VIDIOC_REQBUFS 0xc0145608
#define CMD32_VIDIOC_STREAMOFF 0x40045613
#define CMD32_VIDIOC_STREAMON 0x40045612
#define CMD32_VIDIOC_SUBDEV_DV_TIMINGS_CAP 0xc0905664
#define CMD32_VIDIOC_SUBDEV_ENUMSTD 0xc0405619
#define CMD32_VIDIOC_SUBDEV_ENUM_DV_TIMINGS 0xc0945662
#define CMD32_VIDIOC_SUBDEV_ENUM_FRAME_INTERVAL 0xc040564b
#define CMD32_VIDIOC_SUBDEV_ENUM_FRAME_SIZE 0xc040564a
#define CMD32_VIDIOC_SUBDEV_ENUM_MBUS_CODE 0xc0305602
#define CMD32_VIDIOC_SUBDEV_G_CLIENT_CAP 0x80085665
#define CMD32_VIDIOC_SUBDEV_G_CROP 0xc038563b
#define CMD32_VIDIOC_SUBDEV_G_DV_TIMINGS 0xc0845658
#define CMD32_VIDIOC_SUBDEV_G_EDID 0xc0245628
#define CMD32_VIDIOC_SUBDEV_G_FMT 0xc0585604
#define CMD32_VIDIOC_SUBDEV_G_FRAME_INTERVAL 0xc0305615
#define CMD32_VIDIOC_SUBDEV_G_ROUTING 0xc0405626
#define CMD32_VIDIOC_SUBDEV_G_SELECTION 0xc040563d
#define CMD32_VIDIOC_SUBDEV_G_STD 0x80085617
#define CMD32_VIDIOC_SUBDEV_QUERYCAP 0x80405600
#define CMD32_VIDIOC_SUBDEV_QUERYSTD 0x8008563f
#define CMD32_VIDIOC_SUBDEV_QUERY_DV_TIMINGS 0x80845663
#define CMD32_VIDIOC_SUBDEV_S_CLIENT_CAP 0xc0085666
#define CMD32_VIDIOC_SUBDEV_S_CROP 0xc038563c
#define CMD32_VIDIOC_SUBDEV_S_DV_TIMINGS 0xc0845657
#define CMD32_VIDIOC_SUBDEV_S_EDID 0xc0245629
#define CMD32_VIDIOC_SUBDEV_S_FMT 0xc0585605
#define CMD32_VIDIOC_SUBDEV_S_FRAME_INTERVAL 0xc0305616
#define CMD32_VIDIOC_SUBDEV_S_ROUTING 0xc0405627
#define CMD32_VIDIOC_SUBDEV_S_SELECTION 0xc040563e
#define CMD32_VIDIOC_SUBDEV_S_STD 0x40085618
#define CMD32_VIDIOC_SUBSCRIBE_EVENT 0x4020565a
#define CMD32_VIDIOC_S_AUDIO 0x40345622
#define CMD32_VIDIOC_S_AUDOUT 0x40345632
#define CMD32_VIDIOC_S_CROP 0x4014563c
#define CMD32_VIDIOC_S_CTRL 0xc008561c
#define CMD32_VIDIOC_S_DV_TIMINGS 0xc0845657
#define CMD32_VIDIOC_S_EDID 0xc0245629
#define CMD32_VIDIOC_S_EXT_CTRLS 0xc0185648
#define CMD32_VIDIOC_S_FBUF 0x402c560b
#define CMD32_VIDIOC_S_FMT 0xc0cc5605
#define CMD32_VIDIOC_S_FREQUENCY 0x402c5639
#define CMD32_VIDIOC_S_HW_FREQ_SEEK 0x40305652
#define CMD32_VIDIOC_S_INPUT 0xc0045627
#define CMD32_VIDIOC_S_JPEGCOMP 0x408c563e
#define CMD32_VIDIOC_S_MODULATOR 0x40445637
#define CMD32_VIDIOC_S_OUTPUT 0xc004562f
#define CMD32_VIDIOC_S_PARM 0xc0cc5616
#define CMD32_VIDIOC_S_PRIORITY 0x40045644
#define CMD32_VIDIOC_S_SELECTION 0xc040565f
#define CMD32_VIDIOC_S_STD 0x40085618
#define CMD32_VIDIOC_S_TUNER 0x4054561e
#define CMD32_VIDIOC_TRY_DECODER_CMD 0xc0485661
#define CMD32_VIDIOC_TRY_ENCODER_CMD 0xc028564e
#define CMD32_VIDIOC_TRY_EXT_CTRLS 0xc0185649
#define CMD32_VIDIOC_TRY_FMT 0xc0cc5640
#define CMD32_VIDIOC_UNSUBSCRIBE_EVENT 0x4020565b
